/* Haar Filter
 * Copyright (c) 2002-2003 Emil Mikulic.
 * http://dmr.ath.cx/
 *
 * Feel free to do whatever you like with this code.
 * Feel free to credit me.
 *
 * $Id: filter.c,v 1.3 2004/10/09 09:25:05 emikulic Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "targa.h" /* look at the URL given above */

typedef double val;
typedef uint8_t byte;

int image_size = 0;



static void *xmalloc(const size_t amount)
{
	void *ptr = malloc(amount);
	if (!ptr)
	{
		printf("Ran out of memory in xmalloc()\n");
		exit(EXIT_FAILURE);
	}
	return ptr;
}



static void do_haar(val *array)
{
	val *avg, *diff;
	int i, count;

	avg = (val*)xmalloc(sizeof(val) * image_size);
	diff = (val*)xmalloc(sizeof(val) * image_size);

	for (count=image_size; count>1; count/=2)
	{
		for (i=0; i<count/2; i++)
		{
			avg[i] = (array[2*i] + array[2*i + 1]) / 2;
			diff[i] = array[2*i] - avg[i];
		}

		for (i=0; i<count/2; i++)
		{
			array[i] = avg[i];
			array[i + count/2] = diff[i];
		}
	}

	free(diff);
	free(avg);
}



#define DONTFAIL(x) do { tga_result res;  if ((res=x) != TGA_NOERR) { \
	printf("Targa error: %s\n", tga_error(res)); \
	exit(EXIT_FAILURE); } } while(0)

static val* load_targa(const char *fn)
{
	tga_image tga;
	int i,j;
	byte *ptr;
	val *dest;

	DONTFAIL( tga_read(&tga, fn) );

	printf("Loaded %dx%dx%dbpp targa (\"%s\").\n",
		tga.width, tga.height, tga.pixel_depth, fn);

	if (!tga_is_mono(&tga)) DONTFAIL( tga_desaturate_rec_601_1(&tga) );
	if (!tga_is_top_to_bottom(&tga)) DONTFAIL( tga_flip_vert(&tga) );
	if (tga_is_right_to_left(&tga)) DONTFAIL( tga_flip_horiz(&tga) );

	if (tga.width != tga.height)
	{
		printf("Targa must be square!\n");
		exit(EXIT_FAILURE);
	}
	if (image_size && tga.width != image_size)
	{
		printf("Targa must be %dx%d\n", image_size, image_size);
		exit(EXIT_FAILURE);
	}
	if (!image_size)
	{
		int shift = 0;
		image_size = tga.width;
		shift = (int)( log((val)image_size) / log(2.0) );
		if (1 << shift != image_size)
		{
			printf("Resolution %d is not a power of two!\n",
				image_size);
			exit(EXIT_FAILURE);
		}
	}

	/* image_data -> dest */
	ptr = tga.image_data;
	dest = (val*)xmalloc(sizeof(val) * image_size*image_size);

	for (j=0; j<image_size; j++)
	for (i=0; i<image_size; i++)
		dest[j * image_size + i] = (val)(*ptr++) / 255.0;

	tga_free_buffers(&tga);
	return dest;
}



static void transform_row(val *data, const int row)
{
	do_haar(&data[row*image_size]);
}



static void transform_col(val *data, const int col)
{
	val *coldata;
	int i;

	coldata = (val*)xmalloc(sizeof(val) * image_size);
	for (i=0; i<image_size; i++)
		coldata[i] = data[i*image_size + col];

	do_haar(coldata);

	for (i=0; i<image_size; i++)
		data[i*image_size + col] = coldata[i];
	free(coldata);
}



static void write_haar_img(const val *data, const char *fn)
{
	uint8_t *img;
	val max_neg, max_pos;
	int i,j;

	img = (uint8_t*)xmalloc(image_size*image_size*3);

	max_neg = max_pos = 0.0;

	for (j=0; j<image_size; j++)
	for (i=0; i<image_size; i++)
	{
		if (data[j*image_size+i] < max_neg)
			max_neg = data[j*image_size+i];
		if (data[j*image_size+i] > max_pos)
			max_pos = data[j*image_size+i];
	}

	for (j=0; j<image_size; j++)
	for (i=0; i<image_size; i++)
	{
		byte b;
		val v, orig_v = data[j*image_size + i];

		v = orig_v;
		if (v < 0)
			v /= max_neg;
		else
			v /= max_pos;

		v = sqrt(v);
		b = 255 - (byte)(v * 255.0);

		if (orig_v < 0)
		{
			img[(j*image_size + i)*3+0] = b;
			img[(j*image_size + i)*3+1] = b;
			img[(j*image_size + i)*3+2] = 255;
		}
		else
		{
			img[(j*image_size + i)*3+0] = 255;
			img[(j*image_size + i)*3+1] = b;
			img[(j*image_size + i)*3+2] = b;
		}
	}

	DONTFAIL( tga_write_rgb(fn, img, image_size, image_size, 24) );
	free(img);
	printf("wrote \"%s\"\n", fn);
}



static void write_gray_img(const val *data, const char *fn)
{
	uint8_t *img;
	int i,j;

	img = (uint8_t*)xmalloc(image_size*image_size);

	for (j=0; j<image_size; j++)
	for (i=0; i<image_size; i++)
	{
		val v = data[j*image_size + i];
		byte b = (v < 0.0) ? 0 :
			(v > 1.0) ? 255 :
			(byte)(v * 255.0);

		img[j*image_size + i] = b;
	}

	DONTFAIL( tga_write_mono(fn, img, image_size, image_size) );
	free(img);
	printf("wrote \"%s\"\n", fn);
}



static void apply_mask_trans(val *mask, const val trans)
{
	int i;
	for (i=0; i<image_size*image_size; i++)
		mask[i] = (1.0-trans) + trans*mask[i];
}



static void do_mask(val *array, const val mult)
{
	val *avg;
	int i, count;

	avg = (val*)xmalloc(sizeof(val) * image_size);

	for (count=image_size; count>1; count/=2)
	{
		for (i=0; i<count/2; i++)
			avg[i] = (array[2*i] + array[2*i + 1]) / 2;

		for (i=0; i<count/2; i++)
		{
			array[i] = avg[i];
			array[i + count/2] = avg[i];
		}
	}
	free(avg);

	/* attenuate using mult */
	for (count=1; count<image_size; count *= 2)
	for (i=count; i<image_size; i++)
		array[i] *= mult;
}



static void mask_row(val *data, const int row, const val mult)
{
	do_mask(&data[row*image_size], mult);
}



static void mask_col(val *data, const int col, const val mult)
{
	val *coldata;
	int i;

	coldata = (val*)xmalloc(sizeof(val) * image_size);
	for (i=0; i<image_size; i++)
		coldata[i] = data[i*image_size + col];

	do_mask(coldata, mult);

	for (i=0; i<image_size; i++)
		data[i*image_size + col] = coldata[i];
	free(coldata);
}



/* returns the percentage of coefficients under <amount> */
static val percent_under(const val *data, const val *mask, const val amount)
{
	int num_thrown = 0, i;

	for (i=0; i<image_size*image_size; i++)
	{
		val v = data[i] * mask[i];
		if (fabs(v) <= amount) num_thrown++;
	}
	return (val)num_thrown / (val)(image_size*image_size);
}



/* throw away weakest <percentage>% of coefficients */
#define MAX_ITER 250
static void quantize(val *data, const val *mask, const val percentage)
{
	val low, high, thresh, loss;
	int i;

	/* find max */
	low = high = 0.0;
	for (i=0; i<image_size*image_size; i++)
		if (fabs(data[i]) > high) high = fabs(data[i]);

	/* binary search */
	for (i=0; i<MAX_ITER; i++)
	{
		thresh = (low+high)/2.0;
		loss = percent_under(data, mask, thresh);

		printf("binary search: "
			"iteration=%4d, thresh=%4f, loss=%3.2f%%   \r",
			i+1, thresh, loss*100.0);
		fflush(stdout);

		if (loss < percentage)
			low = thresh;
		else
			high = thresh;

		if (fabs(loss - percentage) < 0.01) i = MAX_ITER;
		if (fabs(low - high) < 0.0000001) i = MAX_ITER;
	}
	printf("\n");

	/* zero out anything too low */
	for (i=0; i<image_size*image_size; i++)
	{
		val v = data[i] * mask[i];
		if (fabs(v) < thresh) data[i] = 0.0;
	}
}



static void undo_haar(val *array)
{
	val *tmp;
	int i, count;

	tmp = (val*)xmalloc(sizeof(val) * image_size);
	for (count=2; count<=image_size; count*=2)
	{
		for (i=0; i<count/2; i++)
		{
			tmp[2*i]     = array[i] + array[i + count/2];
			tmp[2*i + 1] = array[i] - array[i + count/2];
		}

		for (i=0; i<count; i++) array[i] = tmp[i];
	}
	free(tmp);
}



static void untransform_row(val *data, const int row)
{
	undo_haar(&data[row * image_size]);
}



static void untransform_col(val *data, const int col)
{
	val *coldata;
	int i;

	coldata = (val*)xmalloc(sizeof(val) * image_size);
	for (i=0; i<image_size; i++) coldata[i] = data[i*image_size + col];

	undo_haar(coldata);

	for (i=0; i<image_size; i++) data[i*image_size + col] = coldata[i];
	free(coldata);
}



int main(int argc, char **argv)
{
	val *image, *mask, mult, trans, thresh;
	int i;

	/* usage */
	if (argc != 6)
	{
		printf("usage: %s infile.tga mask.tga mult trans thresh\n\n"
"\tmult attenuates high-frequency coefficients.\n"
"\ttrans defines how much effect the mask has in quantization.\n"
"\tthresh specifies the amount of coefficients to throw away\n"

"All values are decimal, in the range [0-1].\n\n",
		argv[0]);
		return EXIT_FAILURE;
	}

	mult = atof(argv[3]);
	trans = atof(argv[4]);
	thresh = atof(argv[5]);
	printf("mult = %f\ntrans = %f\nthresh = %f\n",
		mult, trans, thresh);

	image = load_targa(argv[1]);
	mask = load_targa(argv[2]);
	apply_mask_trans(mask, trans);

	printf("haar transforming image\n");
	for (i=0; i<image_size; i++) transform_row(image, i);
	for (i=0; i<image_size; i++) transform_col(image, i);
	write_haar_img(image, "output-coeff-full.tga");

	printf("transforming mask\n");
	for (i=0; i<image_size; i++) mask_row(mask, i, mult);
	for (i=0; i<image_size; i++) mask_col(mask, i, mult);
	write_haar_img(mask, "output-mask.tga");

	printf("quantizing\n");
	quantize(image, mask, thresh);
	write_haar_img(image, "output-coeff-quant.tga");

	printf("inverse transform\n");
	for (i=0; i<image_size; i++) untransform_col(image, i);
	for (i=0; i<image_size; i++) untransform_row(image, i);
	write_gray_img(image, "output-unhaar.tga");

	free(image);
	free(mask);
	return 0;
}
